@@ -3,17 +3,19 @@ package database
 import (
 	"context"
 	"database/sql"
+	_ "embed"
 	"errors"
 	"fmt"
 	"math"
 	"strings"
 	"time"
 
-	"git.sr.ht/~emersion/soju/xirc"
 	_ "github.com/lib/pq"
 	"github.com/prometheus/client_golang/prometheus"
 	promcollectors "github.com/prometheus/client_golang/prometheus/collectors"
 	"gopkg.in/irc.v4"
+
+	"codeberg.org/emersion/soju/xirc"
 )
 
 const postgresQueryTimeout = 5 * time.Second
@@ -26,204 +28,8 @@ CREATE TABLE IF NOT EXISTS "Config" (
 );
 `
 
-const postgresSchema = `
-CREATE TABLE "User" (
-	id SERIAL PRIMARY KEY,
-	username VARCHAR(255) NOT NULL UNIQUE,
-	password VARCHAR(255),
-	admin BOOLEAN NOT NULL DEFAULT FALSE,
-	nick VARCHAR(255),
-	realname VARCHAR(255),
-	created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT now(),
-	enabled BOOLEAN NOT NULL DEFAULT TRUE,
-	downstream_interacted_at TIMESTAMP WITH TIME ZONE
-);
-
-CREATE TYPE sasl_mechanism AS ENUM ('PLAIN', 'EXTERNAL');
-
-CREATE TABLE "Network" (
-	id SERIAL PRIMARY KEY,
-	name VARCHAR(255),
-	"user" INTEGER NOT NULL REFERENCES "User"(id) ON DELETE CASCADE,
-	addr VARCHAR(255) NOT NULL,
-	nick VARCHAR(255),
-	username VARCHAR(255),
-	realname VARCHAR(255),
-	certfp TEXT,
-	pass VARCHAR(255),
-	connect_commands VARCHAR(1023),
-	sasl_mechanism sasl_mechanism,
-	sasl_plain_username VARCHAR(255),
-	sasl_plain_password VARCHAR(255),
-	sasl_external_cert BYTEA,
-	sasl_external_key BYTEA,
-	auto_away BOOLEAN NOT NULL DEFAULT TRUE,
-	enabled BOOLEAN NOT NULL DEFAULT TRUE,
-	UNIQUE("user", addr, nick),
-	UNIQUE("user", name)
-);
-
-CREATE TABLE "Channel" (
-	id SERIAL PRIMARY KEY,
-	network INTEGER NOT NULL REFERENCES "Network"(id) ON DELETE CASCADE,
-	name VARCHAR(255) NOT NULL,
-	key VARCHAR(255),
-	detached BOOLEAN NOT NULL DEFAULT FALSE,
-	detached_internal_msgid VARCHAR(255),
-	relay_detached INTEGER NOT NULL DEFAULT 0,
-	reattach_on INTEGER NOT NULL DEFAULT 0,
-	detach_after INTEGER NOT NULL DEFAULT 0,
-	detach_on INTEGER NOT NULL DEFAULT 0,
-	UNIQUE(network, name)
-);
-
-CREATE TABLE "DeliveryReceipt" (
-	id SERIAL PRIMARY KEY,
-	network INTEGER NOT NULL REFERENCES "Network"(id) ON DELETE CASCADE,
-	target VARCHAR(255) NOT NULL,
-	client VARCHAR(255) NOT NULL DEFAULT '',
-	internal_msgid VARCHAR(255) NOT NULL,
-	UNIQUE(network, target, client)
-);
-
-CREATE TABLE "ReadReceipt" (
-	id SERIAL PRIMARY KEY,
-	network INTEGER NOT NULL REFERENCES "Network"(id) ON DELETE CASCADE,
-	target VARCHAR(255) NOT NULL,
-	timestamp TIMESTAMP WITH TIME ZONE NOT NULL,
-	UNIQUE(network, target)
-);
-
-CREATE TABLE "WebPushConfig" (
-	id SERIAL PRIMARY KEY,
-	created_at TIMESTAMP WITH TIME ZONE NOT NULL,
-	vapid_key_public TEXT NOT NULL,
-	vapid_key_private TEXT NOT NULL,
-	UNIQUE(vapid_key_public)
-);
-
-CREATE TABLE "WebPushSubscription" (
-	id SERIAL PRIMARY KEY,
-	created_at TIMESTAMP WITH TIME ZONE NOT NULL,
-	updated_at TIMESTAMP WITH TIME ZONE NOT NULL,
-	"user" INTEGER NOT NULL REFERENCES "User"(id) ON DELETE CASCADE,
-	network INTEGER REFERENCES "Network"(id) ON DELETE CASCADE,
-	endpoint TEXT NOT NULL,
-	key_vapid TEXT,
-	key_auth TEXT,
-	key_p256dh TEXT,
-	UNIQUE(network, endpoint)
-);
-
-CREATE TABLE "MessageTarget" (
-	id SERIAL PRIMARY KEY,
-	network INTEGER NOT NULL REFERENCES "Network"(id) ON DELETE CASCADE,
-	target TEXT NOT NULL,
-	UNIQUE(network, target)
-);
-
-CREATE TEXT SEARCH DICTIONARY search_simple_dictionary (
-    TEMPLATE = pg_catalog.simple
-);
-CREATE TEXT SEARCH CONFIGURATION @SCHEMA_PREFIX@search_simple ( COPY = pg_catalog.simple );
-ALTER TEXT SEARCH CONFIGURATION @SCHEMA_PREFIX@search_simple ALTER MAPPING FOR asciiword, asciihword, hword_asciipart, hword, hword_part, word WITH @SCHEMA_PREFIX@search_simple_dictionary;
-CREATE TABLE "Message" (
-	id SERIAL PRIMARY KEY,
-	target INTEGER NOT NULL REFERENCES "MessageTarget"(id) ON DELETE CASCADE,
-	raw TEXT NOT NULL,
-	time TIMESTAMP WITH TIME ZONE NOT NULL,
-	sender TEXT NOT NULL,
-	text TEXT,
-	text_search tsvector GENERATED ALWAYS AS (to_tsvector('@SCHEMA_PREFIX@search_simple', text)) STORED
-);
-CREATE INDEX "MessageIndex" ON "Message" (target, time);
-CREATE INDEX "MessageSearchIndex" ON "Message" USING GIN (text_search);
-`
-
-var postgresMigrations = []string{
-	"", // migration #0 is reserved for schema initialization
-	`ALTER TABLE "Network" ALTER COLUMN nick DROP NOT NULL`,
-	`
-		CREATE TYPE sasl_mechanism AS ENUM ('PLAIN', 'EXTERNAL');
-		ALTER TABLE "Network"
-			ALTER COLUMN sasl_mechanism
-			TYPE sasl_mechanism
-			USING sasl_mechanism::sasl_mechanism;
-	`,
-	`
-		CREATE TABLE "ReadReceipt" (
-			id SERIAL PRIMARY KEY,
-			network INTEGER NOT NULL REFERENCES "Network"(id) ON DELETE CASCADE,
-			target VARCHAR(255) NOT NULL,
-			timestamp TIMESTAMP WITH TIME ZONE NOT NULL,
-			UNIQUE(network, target)
-		);
-	`,
-	`
-		CREATE TABLE "WebPushConfig" (
-			id SERIAL PRIMARY KEY,
-			created_at TIMESTAMP WITH TIME ZONE NOT NULL,
-			vapid_key_public TEXT NOT NULL,
-			vapid_key_private TEXT NOT NULL,
-			UNIQUE(vapid_key_public)
-		);
-
-		CREATE TABLE "WebPushSubscription" (
-			id SERIAL PRIMARY KEY,
-			created_at TIMESTAMP WITH TIME ZONE NOT NULL,
-			updated_at TIMESTAMP WITH TIME ZONE NOT NULL,
-			network INTEGER REFERENCES "Network"(id) ON DELETE CASCADE,
-			endpoint TEXT NOT NULL,
-			key_vapid TEXT,
-			key_auth TEXT,
-			key_p256dh TEXT,
-			UNIQUE(network, endpoint)
-		);
-	`,
-	`
-		ALTER TABLE "WebPushSubscription"
-		ADD COLUMN "user" INTEGER
-		REFERENCES "User"(id) ON DELETE CASCADE
-	`,
-	`ALTER TABLE "User" ADD COLUMN nick VARCHAR(255)`,
-	// Before this migration, a bug swapped user and network, so empty the
-	// web push subscriptions table
-	`
-		DELETE FROM "WebPushSubscription";
-		ALTER TABLE "WebPushSubscription"
-		ALTER COLUMN "user"
-		SET NOT NULL;
-	`,
-	`ALTER TABLE "Network" ADD COLUMN auto_away BOOLEAN NOT NULL DEFAULT TRUE`,
-	`ALTER TABLE "Network" ADD COLUMN certfp TEXT`,
-	`ALTER TABLE "User" ADD COLUMN created_at TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT now()`,
-	`ALTER TABLE "User" ADD COLUMN enabled BOOLEAN NOT NULL DEFAULT TRUE`,
-	`ALTER TABLE "User" ADD COLUMN downstream_interacted_at TIMESTAMP WITH TIME ZONE`,
-	`
-		CREATE TABLE "MessageTarget" (
-			id SERIAL PRIMARY KEY,
-			network INTEGER NOT NULL REFERENCES "Network"(id) ON DELETE CASCADE,
-			target TEXT NOT NULL,
-			UNIQUE(network, target)
-		);
-		CREATE TEXT SEARCH DICTIONARY search_simple_dictionary (
-			TEMPLATE = pg_catalog.simple
-		);
-		CREATE TEXT SEARCH CONFIGURATION @SCHEMA_PREFIX@search_simple ( COPY = pg_catalog.simple );
-		ALTER TEXT SEARCH CONFIGURATION @SCHEMA_PREFIX@search_simple ALTER MAPPING FOR asciiword, asciihword, hword_asciipart, hword, hword_part, word WITH @SCHEMA_PREFIX@search_simple_dictionary;
-		CREATE TABLE "Message" (
-			id SERIAL PRIMARY KEY,
-			target INTEGER NOT NULL REFERENCES "MessageTarget"(id) ON DELETE CASCADE,
-			raw TEXT NOT NULL,
-			time TIMESTAMP WITH TIME ZONE NOT NULL,
-			sender TEXT NOT NULL,
-			text TEXT,
-			text_search tsvector GENERATED ALWAYS AS (to_tsvector('@SCHEMA_PREFIX@search_simple', text)) STORED
-		);
-		CREATE INDEX "MessageIndex" ON "Message" (target, time);
-		CREATE INDEX "MessageSearchIndex" ON "Message" USING GIN (text_search);
-	`,
-}
+//go:embed postgres_schema.sql
+var postgresSchema string
 
 type PostgresDB struct {
 	db   *sql.DB
@@ -367,7 +173,7 @@ func (db *PostgresDB) ListUsers(ctx context.Context) ([]User, error) {
 
 	rows, err := db.db.QueryContext(ctx,
 		`SELECT id, username, password, admin, nick, realname, enabled,
-			downstream_interacted_at
+			downstream_interacted_at, max_networks
 		FROM "User"`)
 	if err != nil {
 		return nil, err
@@ -379,7 +185,7 @@ func (db *PostgresDB) ListUsers(ctx context.Context) ([]User, error) {
 		var user User
 		var password, nick, realname sql.NullString
 		var downstreamInteractedAt sql.NullTime
-		if err := rows.Scan(&user.ID, &user.Username, &password, &user.Admin, &nick, &realname, &user.Enabled, &downstreamInteractedAt); err != nil {
+		if err := rows.Scan(&user.ID, &user.Username, &password, &user.Admin, &nick, &realname, &user.Enabled, &downstreamInteractedAt, &user.MaxNetworks); err != nil {
 			return nil, err
 		}
 		user.Password = password.String
@@ -404,11 +210,11 @@ func (db *PostgresDB) GetUser(ctx context.Context, username string) (*User, erro
 	var password, nick, realname sql.NullString
 	var downstreamInteractedAt sql.NullTime
 	row := db.db.QueryRowContext(ctx,
-		`SELECT id, password, admin, nick, realname, enabled, downstream_interacted_at
+		`SELECT id, password, admin, nick, realname, enabled, downstream_interacted_at, max_networks
 		FROM "User"
 		WHERE username = $1`,
 		username)
-	if err := row.Scan(&user.ID, &password, &user.Admin, &nick, &realname, &user.Enabled, &downstreamInteractedAt); err != nil {
+	if err := row.Scan(&user.ID, &password, &user.Admin, &nick, &realname, &user.Enabled, &downstreamInteractedAt, &user.MaxNetworks); err != nil {
 		return nil, err
 	}
 	user.Password = password.String
@@ -458,19 +264,19 @@ func (db *PostgresDB) StoreUser(ctx context.Context, user *User) error {
 	if user.ID == 0 {
 		err = db.db.QueryRowContext(ctx, `
 			INSERT INTO "User" (username, password, admin, nick, realname,
-				enabled, downstream_interacted_at)
-			VALUES ($1, $2, $3, $4, $5, $6, $7)
+				enabled, downstream_interacted_at, max_networks)
+			VALUES ($1, $2, $3, $4, $5, $6, $7, $8)
 			RETURNING id`,
 			user.Username, password, user.Admin, nick, realname, user.Enabled,
-			downstreamInteractedAt).Scan(&user.ID)
+			downstreamInteractedAt, user.MaxNetworks).Scan(&user.ID)
 	} else {
 		_, err = db.db.ExecContext(ctx, `
 			UPDATE "User"
 			SET password = $1, admin = $2, nick = $3, realname = $4,
-				enabled = $5, downstream_interacted_at = $6
-			WHERE id = $7`,
+				enabled = $5, downstream_interacted_at = $6, max_networks = $7
+			WHERE id = $8`,
 			password, user.Admin, nick, realname, user.Enabled,
-			downstreamInteractedAt, user.ID)
+			downstreamInteractedAt, user.MaxNetworks, user.ID)
 	}
 	return err
 }
@@ -913,9 +719,12 @@ func (db *PostgresDB) GetMessageLastID(ctx context.Context, networkID int64, nam
 
 	var msgID int64
 	row := db.db.QueryRowContext(ctx, `
-		SELECT m.id FROM "Message" AS m, "MessageTarget" as t
-		WHERE t.network = $1 AND t.target = $2 AND m.target = t.id
-		ORDER BY m.time DESC LIMIT 1`,
+		SELECT id FROM "Message"
+		WHERE target = (
+			SELECT id FROM "MessageTarget"
+			WHERE network = $1 AND target = $2
+		)
+		ORDER BY time DESC LIMIT 1`,
 		networkID,
 		name,
 	)
@@ -928,31 +737,88 @@ func (db *PostgresDB) GetMessageLastID(ctx context.Context, networkID int64, nam
 	return msgID, nil
 }
 
-func (db *PostgresDB) StoreMessage(ctx context.Context, networkID int64, name string, msg *irc.Message) (int64, error) {
+func (db *PostgresDB) GetMessageTarget(ctx context.Context, networkID int64, target string) (*MessageTarget, error) {
 	ctx, cancel := context.WithTimeout(ctx, postgresQueryTimeout)
 	defer cancel()
 
-	var t time.Time
-	if tag, ok := msg.Tags["time"]; ok {
-		var err error
-		t, err = time.Parse(xirc.ServerTimeLayout, tag)
+	mt := &MessageTarget{
+		Target: target,
+	}
+
+	row := db.db.QueryRowContext(ctx,
+		`SELECT id, pinned, muted, blocked FROM "MessageTarget" WHERE network = $1 AND target = $2`,
+		networkID, target)
+	if err := row.Scan(&mt.ID, &mt.Pinned, &mt.Muted, &mt.Blocked); err != nil && err != sql.ErrNoRows {
+		return nil, err
+	}
+	return mt, nil
+}
+
+func (db *PostgresDB) ListMessageTargets(ctx context.Context, networkID int64) ([]MessageTarget, error) {
+	ctx, cancel := context.WithTimeout(ctx, postgresQueryTimeout)
+	defer cancel()
+
+	rows, err := db.db.QueryContext(ctx, `
+		SELECT id, target, pinned, muted, blocked
+		FROM "MessageTarget"
+		WHERE network = $1 AND (pinned OR muted OR blocked)`, networkID)
+	if err != nil {
+		return nil, err
+	}
+	defer rows.Close()
+
+	var mts []MessageTarget
+	for rows.Next() {
+		var mt MessageTarget
+		err := rows.Scan(&mt.ID, &mt.Target, &mt.Pinned, &mt.Muted, &mt.Blocked)
 		if err != nil {
-			return 0, fmt.Errorf("failed to parse message time tag: %v", err)
+			return nil, err
 		}
+		mts = append(mts, mt)
+	}
+	if err := rows.Err(); err != nil {
+		return nil, err
+	}
+
+	return mts, nil
+}
+
+func (db *PostgresDB) StoreMessageTarget(ctx context.Context, networkID int64, mt *MessageTarget) error {
+	ctx, cancel := context.WithTimeout(ctx, postgresQueryTimeout)
+	defer cancel()
+
+	var err error
+	if mt.ID != 0 {
+		_, err = db.db.ExecContext(ctx, `
+			UPDATE "MessageTarget"
+			SET pinned = $1, muted = $2, blocked = $3
+			WHERE id = $4`,
+			mt.Pinned, mt.Muted, mt.Blocked, mt.ID)
 	} else {
-		t = time.Now()
+		err = db.db.QueryRowContext(ctx, `
+			INSERT INTO "MessageTarget" (network, target, pinned, muted, blocked)
+			VALUES ($1, $2, $3, $4, $5)
+			RETURNING id`,
+			networkID, mt.Target, mt.Pinned, mt.Muted, mt.Blocked).Scan(&mt.ID)
+	}
+	return err
+}
+
+func (db *PostgresDB) StoreMessages(ctx context.Context, networkID int64, name string, msgs []*irc.Message) ([]int64, error) {
+	if len(msgs) == 0 {
+		return nil, nil
 	}
 
-	var text sql.NullString
-	switch msg.Command {
-	case "PRIVMSG", "NOTICE":
-		if len(msg.Params) > 1 {
-			text.Valid = true
-			text.String = msg.Params[1]
-		}
-	}
+	ctx, cancel := context.WithTimeout(ctx, time.Duration(len(msgs))*postgresQueryTimeout)
+	defer cancel()
 
-	_, err := db.db.ExecContext(ctx, `
+	tx, err := db.db.BeginTx(ctx, nil)
+	if err != nil {
+		return nil, err
+	}
+	defer tx.Rollback()
+
+	_, err = tx.ExecContext(ctx, `
 		INSERT INTO "MessageTarget" (network, target)
 		VALUES ($1, $2)
 		ON CONFLICT DO NOTHING`,
@@ -960,30 +826,59 @@ func (db *PostgresDB) StoreMessage(ctx context.Context, networkID int64, name st
 		name,
 	)
 	if err != nil {
-		return 0, err
+		return nil, err
 	}
 
-	var id int64
-	err = db.db.QueryRowContext(ctx, `
+	insertStmt, err := tx.PrepareContext(ctx, `
 		INSERT INTO "Message" (target, raw, time, sender, text)
 		SELECT id, $1, $2, $3, $4
 		FROM "MessageTarget" as t
 		WHERE network = $5 AND target = $6
-		RETURNING id`,
-		msg.String(),
-		t,
-		msg.Name,
-		text,
-		networkID,
-		name,
-	).Scan(&id)
+		RETURNING id`)
 	if err != nil {
-		return 0, err
+		return nil, err
 	}
-	return id, nil
+
+	ids := make([]int64, len(msgs))
+	for i, msg := range msgs {
+		var t time.Time
+		if tag, ok := msg.Tags["time"]; ok {
+			var err error
+			t, err = time.Parse(xirc.ServerTimeLayout, tag)
+			if err != nil {
+				return nil, fmt.Errorf("failed to parse message time tag: %w", err)
+			}
+		} else {
+			t = time.Now()
+		}
+
+		var text sql.NullString
+		switch msg.Command {
+		case "PRIVMSG", "NOTICE":
+			if len(msg.Params) > 1 {
+				text.Valid = true
+				text.String = stripANSI(msg.Params[1])
+			}
+		}
+
+		err = insertStmt.QueryRowContext(ctx,
+			msg.String(),
+			t,
+			msg.Name,
+			text,
+			networkID,
+			name,
+		).Scan(&ids[i])
+		if err != nil {
+			return nil, err
+		}
+	}
+
+	err = tx.Commit()
+	return ids, err
 }
 
-func (db *PostgresDB) ListMessageLastPerTarget(ctx context.Context, networkID int64, options *MessageOptions) ([]MessageTarget, error) {
+func (db *PostgresDB) ListMessageLastPerTarget(ctx context.Context, networkID int64, options *MessageOptions) ([]MessageTargetLast, error) {
 	ctx, cancel := context.WithTimeout(ctx, postgresQueryTimeout)
 	defer cancel()
 
@@ -991,31 +886,38 @@ func (db *PostgresDB) ListMessageLastPerTarget(ctx context.Context, networkID in
 		networkID,
 	}
 	query := `
-		SELECT t.target, MAX(m.time) AS latest
-		FROM "Message" m, "MessageTarget" t
-		WHERE m.target = t.id AND t.network = $1
+		SELECT t.target, l.latest
+		FROM "MessageTarget" t JOIN LATERAL (
+			SELECT m.target, m.time AS latest, m.text
+			FROM "Message" m
+			WHERE m.target = t.id
 	`
+
 	if !options.Events {
 		query += `AND m.text IS NOT NULL `
 	}
+
 	query += `
-		GROUP BY t.target
-		HAVING true
+
+			ORDER BY m.time DESC LIMIT 1
+		) AS l ON t.id = l.target
+		WHERE t.network = $1
 	`
+
 	if !options.AfterTime.IsZero() {
 		// compares time strings by lexicographical order
 		parameters = append(parameters, options.AfterTime)
-		query += fmt.Sprintf(`AND MAX(m.time) > $%d `, len(parameters))
+		query += fmt.Sprintf(`AND l.latest > $%d `, len(parameters))
 	}
 	if !options.BeforeTime.IsZero() {
 		// compares time strings by lexicographical order
 		parameters = append(parameters, options.BeforeTime)
-		query += fmt.Sprintf(`AND MAX(m.time) < $%d `, len(parameters))
+		query += fmt.Sprintf(`AND l.latest < $%d `, len(parameters))
 	}
 	if options.TakeLast {
-		query += `ORDER BY latest DESC `
+		query += `ORDER BY l.latest DESC `
 	} else {
-		query += `ORDER BY latest ASC `
+		query += `ORDER BY l.latest ASC `
 	}
 	parameters = append(parameters, options.Limit)
 	query += fmt.Sprintf(`LIMIT $%d`, len(parameters))
@@ -1026,9 +928,9 @@ func (db *PostgresDB) ListMessageLastPerTarget(ctx context.Context, networkID in
 	}
 	defer rows.Close()
 
-	var l []MessageTarget
+	var l []MessageTargetLast
 	for rows.Next() {
-		var mt MessageTarget
+		var mt MessageTargetLast
 		if err := rows.Scan(&mt.Name, &mt.LatestMessage); err != nil {
 			return nil, err
 		}
@@ -1059,38 +961,40 @@ func (db *PostgresDB) ListMessages(ctx context.Context, networkID int64, name st
 		name,
 	}
 	query := `
-		SELECT m.raw
-		FROM "Message" AS m, "MessageTarget" AS t
-		WHERE m.target = t.id AND t.network = $1 AND t.target = $2 `
+		SELECT raw FROM "Message"
+		WHERE target = (
+			SELECT id FROM "MessageTarget"
+			WHERE network = $1 AND target = $2
+		) `
 	if options.AfterID > 0 {
 		parameters = append(parameters, options.AfterID)
-		query += fmt.Sprintf(`AND m.id > $%d `, len(parameters))
+		query += fmt.Sprintf(`AND id > $%d `, len(parameters))
 	}
 	if !options.AfterTime.IsZero() {
 		// compares time strings by lexicographical order
 		parameters = append(parameters, options.AfterTime)
-		query += fmt.Sprintf(`AND m.time > $%d `, len(parameters))
+		query += fmt.Sprintf(`AND time > $%d `, len(parameters))
 	}
 	if !options.BeforeTime.IsZero() {
 		// compares time strings by lexicographical order
 		parameters = append(parameters, options.BeforeTime)
-		query += fmt.Sprintf(`AND m.time < $%d `, len(parameters))
+		query += fmt.Sprintf(`AND time < $%d `, len(parameters))
 	}
 	if options.Sender != "" {
 		parameters = append(parameters, options.Sender)
-		query += fmt.Sprintf(`AND m.sender = $%d `, len(parameters))
+		query += fmt.Sprintf(`AND sender = $%d `, len(parameters))
 	}
 	if options.Text != "" {
 		parameters = append(parameters, options.Text)
 		query += fmt.Sprintf(`AND text_search @@ plainto_tsquery('search_simple', $%d) `, len(parameters))
 	}
 	if !options.Events {
-		query += `AND m.text IS NOT NULL `
+		query += `AND text IS NOT NULL `
 	}
 	if options.TakeLast {
-		query += `ORDER BY m.time DESC `
+		query += `ORDER BY time DESC `
 	} else {
-		query += `ORDER BY m.time ASC `
+		query += `ORDER BY time ASC `
 	}
 	parameters = append(parameters, options.Limit)
 	query += fmt.Sprintf(`LIMIT $%d`, len(parameters))
