@@ -12,6 +12,7 @@ import (
 	"errors"
 	"fmt"
 	"io"
+	"math/rand"
 	"net"
 	"strconv"
 	"strings"
@@ -20,10 +21,12 @@ import (
 	"github.com/emersion/go-sasl"
 	"gopkg.in/irc.v4"
 
-	"git.sr.ht/~emersion/soju/database"
-	"git.sr.ht/~emersion/soju/xirc"
+	"codeberg.org/emersion/soju/database"
+	"codeberg.org/emersion/soju/xirc"
 )
 
+const contextSourceBlockedKey = "source-blocked"
+
 // permanentUpstreamCaps is the static list of upstream capabilities always
 // requested when supported.
 var permanentUpstreamCaps = map[string]bool{
@@ -44,12 +47,14 @@ var permanentUpstreamCaps = map[string]bool{
 
 	"draft/account-registration": true,
 	"draft/extended-monitor":     true,
+	"draft/message-redaction":    true,
 }
 
 // storableMessageTags is the static list of message tags that will cause
 // a TAGMSG to be stored.
 var storableMessageTags = map[string]bool{
-	"+react": true,
+	"+draft/react": true,
+	"+react":       true,
 }
 
 type registrationError struct {
@@ -189,7 +194,7 @@ type pendingUpstreamCommand struct {
 }
 
 type upstreamConn struct {
-	conn
+	*conn
 
 	network *network
 	user    *user
@@ -199,6 +204,7 @@ type upstreamConn struct {
 	availableUserModes    string
 	availableChannelModes map[byte]channelModeType
 	availableChannelTypes string
+	availableStatusMsg    string
 	availableMemberships  []xirc.Membership
 	isupport              map[string]*string
 
@@ -240,8 +246,6 @@ func connectToUpstream(ctx context.Context, network *network) (*upstreamConn, er
 	ctx, cancel := context.WithTimeout(ctx, connectTimeout)
 	defer cancel()
 
-	var dialer net.Dialer
-
 	u, err := network.URL()
 	if err != nil {
 		return nil, err
@@ -257,13 +261,6 @@ func connectToUpstream(ctx context.Context, network *network) (*upstreamConn, er
 			addr = u.Host + ":6697"
 		}
 
-		dialer.LocalAddr, err = network.user.localTCPAddrForHost(ctx, host)
-		if err != nil {
-			return nil, fmt.Errorf("failed to pick local IP for remote host %q: %v", host, err)
-		}
-
-		logger.Printf("connecting to TLS server at address %q", addr)
-
 		tlsConfig := &tls.Config{ServerName: host, NextProtos: []string{"irc"}}
 		if network.SASL.Mechanism == "EXTERNAL" {
 			if network.SASL.External.CertBlob == nil {
@@ -319,9 +316,10 @@ func connectToUpstream(ctx context.Context, network *network) (*upstreamConn, er
 			}
 		}
 
-		netConn, err = dialer.DialContext(ctx, "tcp", addr)
+		logger.Printf("connecting to TLS server at address %q", addr)
+		netConn, err = dialTCP(ctx, network.user, addr)
 		if err != nil {
-			return nil, fmt.Errorf("failed to dial %q: %v", addr, err)
+			return nil, err
 		}
 
 		// Don't do the TLS handshake immediately, because we need to register
@@ -330,23 +328,17 @@ func connectToUpstream(ctx context.Context, network *network) (*upstreamConn, er
 		netConn = tls.Client(netConn, tlsConfig)
 	case "irc+insecure":
 		addr := u.Host
-		host, _, err := net.SplitHostPort(addr)
-		if err != nil {
-			host = u.Host
+		if _, _, err := net.SplitHostPort(addr); err != nil {
 			addr = u.Host + ":6667"
 		}
 
-		dialer.LocalAddr, err = network.user.localTCPAddrForHost(ctx, host)
-		if err != nil {
-			return nil, fmt.Errorf("failed to pick local IP for remote host %q: %v", host, err)
-		}
-
 		logger.Printf("connecting to plain-text server at address %q", addr)
-		netConn, err = dialer.DialContext(ctx, "tcp", addr)
+		netConn, err = dialTCP(ctx, network.user, addr)
 		if err != nil {
-			return nil, fmt.Errorf("failed to dial %q: %v", addr, err)
+			return nil, err
 		}
 	case "irc+unix", "unix":
+		var dialer net.Dialer
 		logger.Printf("connecting to Unix socket at path %q", u.Path)
 		netConn, err = dialer.DialContext(ctx, "unix", u.Path)
 		if err != nil {
@@ -364,7 +356,7 @@ func connectToUpstream(ctx context.Context, network *network) (*upstreamConn, er
 
 	cm := stdCaseMapping
 	uc := &upstreamConn{
-		conn:                  *newConn(network.user.srv, newNetIRCConn(netConn), &options),
+		conn:                  newConn(network.user.srv, newNetIRCConn(netConn), &options),
 		network:               network,
 		user:                  network.user,
 		channels:              xirc.NewCaseMappingMap[*upstreamChannel](cm),
@@ -373,6 +365,7 @@ func connectToUpstream(ctx context.Context, network *network) (*upstreamConn, er
 		batches:               make(map[string]upstreamBatch),
 		serverPrefix:          &irc.Prefix{Name: "*"},
 		availableChannelTypes: stdChannelTypes,
+		availableStatusMsg:    "",
 		availableChannelModes: stdChannelModes,
 		availableMemberships:  stdMemberships,
 		isupport:              make(map[string]*string),
@@ -383,6 +376,32 @@ func connectToUpstream(ctx context.Context, network *network) (*upstreamConn, er
 	return uc, nil
 }
 
+func dialTCP(ctx context.Context, user *user, addr string) (net.Conn, error) {
+	var dialer net.Dialer
+	upstreamUserIPs := user.srv.Config().UpstreamUserIPs
+	if len(upstreamUserIPs) > 0 {
+		host, port, err := net.SplitHostPort(addr)
+		if err != nil {
+			return nil, err
+		}
+
+		ipAddr, err := resolveIPAddr(ctx, host)
+		if err != nil {
+			return nil, fmt.Errorf("failed to resolve host %q: %v", host, err)
+		}
+
+		localAddr, err := user.localTCPAddr(ipAddr.IP)
+		if err != nil {
+			return nil, fmt.Errorf("failed to pick local IP for remote host %q: %v", host, err)
+		}
+
+		addr = net.JoinHostPort(ipAddr.String(), port)
+		dialer.LocalAddr = localAddr
+	}
+
+	return dialer.DialContext(ctx, "tcp", addr)
+}
+
 func (uc *upstreamConn) forEachDownstream(f func(*downstreamConn)) {
 	uc.network.forEachDownstream(f)
 }
@@ -421,13 +440,20 @@ func (uc *upstreamConn) isOurNick(nick string) bool {
 	return uc.network.equalCasemap(uc.nick, nick)
 }
 
-func (uc *upstreamConn) forwardMsgByID(id uint64, msg *irc.Message) {
+func (uc *upstreamConn) forwardMessage(ctx context.Context, msg *irc.Message) {
+	uc.forEachDownstream(func(dc *downstreamConn) {
+		dc.SendMessage(ctx, msg)
+	})
+}
+
+func (uc *upstreamConn) forwardMsgByID(ctx context.Context, id uint64, msg *irc.Message) {
 	uc.forEachDownstreamByID(id, func(dc *downstreamConn) {
-		dc.SendMessage(msg)
+		dc.SendMessage(ctx, msg)
 	})
 }
 
 func (uc *upstreamConn) abortPendingCommands() {
+	ctx := context.TODO()
 	for _, l := range uc.pendingCmds {
 		for _, pendingCmd := range l {
 			dc := uc.downstreamByID(pendingCmd.downstreamID)
@@ -437,8 +463,7 @@ func (uc *upstreamConn) abortPendingCommands() {
 
 			switch pendingCmd.msg.Command {
 			case "LIST":
-				dc.SendMessage(&irc.Message{
-					Prefix:  dc.srv.prefix(),
+				dc.SendMessage(ctx, &irc.Message{
 					Command: irc.RPL_LISTEND,
 					Params:  []string{dc.nick, "Command aborted"},
 				})
@@ -447,27 +472,23 @@ func (uc *upstreamConn) abortPendingCommands() {
 				if len(pendingCmd.msg.Params) > 0 {
 					mask = pendingCmd.msg.Params[0]
 				}
-				dc.SendMessage(&irc.Message{
-					Prefix:  dc.srv.prefix(),
+				dc.SendMessage(ctx, &irc.Message{
 					Command: irc.RPL_ENDOFWHO,
 					Params:  []string{dc.nick, mask, "Command aborted"},
 				})
 			case "WHOIS":
 				nick := pendingCmd.msg.Params[len(pendingCmd.msg.Params)-1]
-				dc.SendMessage(&irc.Message{
-					Prefix:  dc.srv.prefix(),
+				dc.SendMessage(ctx, &irc.Message{
 					Command: irc.RPL_ENDOFWHOIS,
 					Params:  []string{dc.nick, nick, "Command aborted"},
 				})
 			case "AUTHENTICATE":
-				dc.endSASL(&irc.Message{
-					Prefix:  dc.srv.prefix(),
+				dc.endSASL(ctx, &irc.Message{
 					Command: irc.ERR_SASLABORTED,
 					Params:  []string{dc.nick, "SASL authentication aborted"},
 				})
 			case "REGISTER", "VERIFY":
-				dc.SendMessage(&irc.Message{
-					Prefix:  dc.srv.prefix(),
+				dc.SendMessage(ctx, &irc.Message{
 					Command: "FAIL",
 					Params:  []string{pendingCmd.msg.Command, "TEMPORARILY_UNAVAILABLE", pendingCmd.msg.Params[0], "Command aborted"},
 				})
@@ -613,9 +634,11 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			Params:  msg.Params,
 		})
 		return nil
-	case "NOTICE", "PRIVMSG", "TAGMSG":
+	case "NOTICE", "PRIVMSG", "TAGMSG", "REDACT":
+		isText := msg.Command == "NOTICE" || msg.Command == "PRIVMSG"
+
 		var target, text string
-		if msg.Command != "TAGMSG" {
+		if isText {
 			if err := parseMessageParams(msg, &target, &text); err != nil {
 				return err
 			}
@@ -625,6 +648,9 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			}
 		}
 
+		// remove statusmsg sigils from target
+		target = strings.TrimLeft(target, uc.availableStatusMsg)
+
 		if uc.network.equalCasemap(msg.Prefix.Name, serviceNick) {
 			uc.logger.Printf("skipping %v from soju's service: %v", msg.Command, msg)
 			break
@@ -636,7 +662,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 
 		if !uc.registered || uc.network.equalCasemap(msg.Prefix.Name, uc.serverPrefix.Name) || target == "*" || strings.HasPrefix(target, "$") {
 			// This is a server message
-			uc.produce("", msg, 0)
+			uc.produce(ctx, "", msg, 0)
 			break
 		}
 
@@ -654,10 +680,10 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 		}
 
 		self := uc.isOurNick(msg.Prefix.Name)
-
 		ch := uc.network.channels.Get(bufferName)
 		highlight := false
-		if ch != nil && msg.Command != "TAGMSG" && !self {
+		detached := false
+		if ch != nil && isText && !self {
 			if ch.Detached {
 				uc.handleDetachedMessage(ctx, ch, msg)
 			}
@@ -666,16 +692,55 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			if ch.DetachOn == database.FilterMessage || ch.DetachOn == database.FilterDefault || (ch.DetachOn == database.FilterHighlight && highlight) {
 				uc.updateChannelAutoDetach(bufferName)
 			}
-		}
 
-		if highlight || directMessage {
-			go uc.network.broadcastWebPush(msg)
-			if timestamp, err := time.Parse(xirc.ServerTimeLayout, string(msg.Tags["time"])); err == nil {
-				uc.network.pushTargets.Set(bufferName, timestamp)
+			if ch.Detached && ch.RelayDetached == database.FilterNone {
+				detached = true
 			}
 		}
 
-		uc.produce(bufferName, msg, downstreamID)
+		sendWebPush := !self && !detached && isText && (highlight || directMessage)
+
+		sourceCM := uc.network.casemap(msg.Prefix.Name)
+		mt, err := uc.srv.db.GetMessageTarget(ctx, uc.network.ID, sourceCM)
+		if err != nil {
+			uc.logger.Printf("failed to get the message target for %q: %v", sourceCM, err)
+		}
+		if mt != nil {
+			if mt.Muted || mt.Blocked {
+				sendWebPush = false
+			}
+			if mt.Blocked {
+				// Used by uc.produce
+				ctx = context.WithValue(ctx, contextSourceBlockedKey, true)
+			}
+		}
+
+		if sendWebPush {
+			timestamp, err := time.Parse(xirc.ServerTimeLayout, msg.Tags["time"])
+			if err != nil {
+				// Cannot fail, was set above if unset or invalid.
+				panic("failed parsing event time")
+			}
+			targetCM := uc.network.casemap(bufferName)
+			time.AfterFunc(5*time.Second, func() {
+				r, err := uc.srv.db.GetReadReceipt(ctx, uc.network.ID, targetCM)
+				if err != nil {
+					uc.logger.Printf("failed to get the read receipt for %q: %v", bufferName, err)
+					return
+				}
+				if r != nil && !r.Timestamp.Before(timestamp) {
+					return
+				}
+
+				uc.network.pushTargetsMutex.Lock()
+				uc.network.pushTargets.Set(bufferName, timestamp)
+				uc.network.pushTargetsMutex.Unlock()
+
+				uc.network.broadcastWebPush(msg)
+			})
+		}
+
+		uc.produce(ctx, bufferName, msg, downstreamID)
 	case "CAP":
 		var subCmd string
 		if err := parseMessageParams(msg, nil, &subCmd); err != nil {
@@ -726,7 +791,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 
 			if uc.registered {
 				uc.forEachDownstream(func(dc *downstreamConn) {
-					dc.updateSupportedCaps()
+					dc.updateSupportedCaps(ctx)
 				})
 			}
 		case "NEW":
@@ -747,11 +812,11 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 
 			if uc.registered {
 				uc.forEachDownstream(func(dc *downstreamConn) {
-					dc.updateSupportedCaps()
+					dc.updateSupportedCaps(ctx)
 				})
 			}
 		default:
-			uc.logger.Printf("unhandled message: %v", msg)
+			uc.logger.Debugf("unhandled message: %v", msg)
 		}
 	case "AUTHENTICATE":
 		if uc.saslClient == nil {
@@ -812,8 +877,8 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 
 		uc.logger.Printf("logged in with account %q", uc.account)
 		uc.forEachDownstream(func(dc *downstreamConn) {
-			dc.updateAccount()
-			dc.updateHost()
+			dc.updateAccount(ctx)
+			dc.updateHost(ctx)
 		})
 	case irc.RPL_LOGGEDOUT:
 		var rawPrefix string
@@ -829,8 +894,8 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 
 		uc.logger.Printf("logged out")
 		uc.forEachDownstream(func(dc *downstreamConn) {
-			dc.updateAccount()
-			dc.updateHost()
+			dc.updateAccount(ctx)
+			dc.updateHost(ctx)
 		})
 	case xirc.RPL_VISIBLEHOST:
 		var rawHost string
@@ -846,7 +911,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 		}
 
 		uc.forEachDownstream(func(dc *downstreamConn) {
-			dc.updateHost()
+			dc.updateHost(ctx)
 		})
 	case irc.ERR_NICKLOCKED, irc.RPL_SASLSUCCESS, irc.ERR_SASLFAIL, irc.ERR_SASLTOOLONG, irc.ERR_SASLABORTED:
 		var info string
@@ -870,7 +935,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 				uc.network.autoSaveSASLPlain(ctx, dc.sasl.plain.Username, dc.sasl.plain.Password)
 			}
 
-			dc.endSASL(msg)
+			dc.endSASL(ctx, msg)
 		}
 
 		if !uc.registered {
@@ -892,7 +957,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 				uc.network.autoSaveSASLPlain(ctx, account, password)
 			}
 
-			dc.SendMessage(msg)
+			dc.SendMessage(ctx, msg)
 		}
 	case irc.RPL_WELCOME:
 		if err := parseMessageParams(msg, &uc.nick); err != nil {
@@ -931,12 +996,16 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			if strings.HasPrefix(token, "-") {
 				negate = true
 				token = token[1:]
-			} else if i := strings.IndexByte(token, '='); i >= 0 {
-				parameter = token[:i]
-				value = token[i+1:]
-				hasValue = true
+			} else {
+				parameter, value, hasValue = strings.Cut(token, "=")
 			}
-			parameter = strings.ToUpper(parameter)
+
+			if prefix, key, ok := strings.Cut(parameter, "/"); ok {
+				parameter = strings.ToLower(prefix) + "/" + strings.ToUpper(key)
+			} else {
+				parameter = strings.ToUpper(parameter)
+			}
+			value = xirc.DecodeIsupportValue(value)
 
 			if hasValue {
 				uc.isupport[parameter] = &value
@@ -966,12 +1035,20 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 				} else {
 					uc.availableChannelTypes = stdChannelTypes
 				}
+			case "STATUSMSG":
+				if !negate {
+					uc.availableStatusMsg = value
+				} else {
+					uc.availableStatusMsg = ""
+				}
 			case "PREFIX":
 				if !negate {
 					err = uc.handleMemberships(value)
 				} else {
 					uc.availableMemberships = stdMemberships
 				}
+			case "soju.im/SAFERATE":
+				uc.rateLimit = negate
 			}
 			if err != nil {
 				return err
@@ -982,12 +1059,12 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			}
 		}
 
-		uc.updateMonitor()
+		uc.updateMonitor(ctx)
 
 		uc.forEachDownstream(func(dc *downstreamConn) {
-			msgs := xirc.GenerateIsupport(dc.srv.prefix(), dc.nick, downstreamIsupport)
+			msgs := xirc.GenerateIsupport(downstreamIsupport)
 			for _, msg := range msgs {
-				dc.SendMessage(msg)
+				dc.SendMessage(ctx, msg)
 			}
 		})
 	case irc.ERR_NOMOTD, irc.RPL_ENDOFMOTD:
@@ -1011,7 +1088,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			return nil
 		}
 
-		uc.forwardMsgByID(downstreamID, msg)
+		uc.forwardMsgByID(ctx, downstreamID, msg)
 	case "BATCH":
 		var tag string
 		if err := parseMessageParams(msg, &tag); err != nil {
@@ -1073,7 +1150,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			if memberships != nil {
 				ch.Members.Del(msg.Prefix.Name)
 				ch.Members.Set(newNick, memberships)
-				uc.appendLog(ch.Name, msg)
+				uc.appendLog(ctx, ch.Name, msg)
 			}
 		})
 
@@ -1082,14 +1159,12 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 		})
 
 		if !me {
-			uc.forEachDownstream(func(dc *downstreamConn) {
-				dc.SendMessage(msg)
-			})
+			uc.forwardMessage(ctx, msg)
 		} else {
 			uc.forEachDownstream(func(dc *downstreamConn) {
-				dc.updateNick()
+				dc.updateNick(ctx)
 			})
-			uc.updateMonitor()
+			uc.updateMonitor(ctx)
 		}
 	case "SETNAME":
 		var newRealname string
@@ -1108,12 +1183,10 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			uc.realname = newRealname
 
 			uc.forEachDownstream(func(dc *downstreamConn) {
-				dc.updateRealname()
+				dc.updateRealname(ctx)
 			})
 		} else {
-			uc.forEachDownstream(func(dc *downstreamConn) {
-				dc.SendMessage(msg)
-			})
+			uc.forwardMessage(ctx, msg)
 		}
 	case "CHGHOST":
 		var newUsername, newHostname string
@@ -1133,13 +1206,11 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			uc.hostname = newHostname
 
 			uc.forEachDownstream(func(dc *downstreamConn) {
-				dc.updateHost()
+				dc.updateHost(ctx)
 			})
 		} else {
-			uc.forEachDownstream(func(dc *downstreamConn) {
-				// TODO: add fallback with QUIT/JOIN/MODE messages
-				dc.SendMessage(msg)
-			})
+			// TODO: add fallback with QUIT/JOIN/MODE messages
+			uc.forwardMessage(ctx, msg)
 		}
 	case "JOIN":
 		var channels string
@@ -1196,7 +1267,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 
 			chMsg := msg.Copy()
 			chMsg.Params[0] = ch
-			uc.produce(ch, chMsg, 0)
+			uc.produce(ctx, ch, chMsg, 0)
 		}
 	case "PART":
 		var channels string
@@ -1229,7 +1300,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 
 			chMsg := msg.Copy()
 			chMsg.Params[0] = ch
-			uc.produce(ch, chMsg, 0)
+			uc.produce(ctx, ch, chMsg, 0)
 		}
 	case "KICK":
 		var channel, user string
@@ -1258,7 +1329,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			}
 		}
 
-		uc.produce(channel, msg, 0)
+		uc.produce(ctx, channel, msg, 0)
 	case "QUIT":
 		if uc.isOurNick(msg.Prefix.Name) {
 			uc.logger.Printf("quit")
@@ -1267,30 +1338,29 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 		uc.channels.ForEach(func(_ string, ch *upstreamChannel) {
 			if ch.Members.Has(msg.Prefix.Name) {
 				ch.Members.Del(msg.Prefix.Name)
-				uc.appendLog(ch.Name, msg)
+				uc.appendLog(ctx, ch.Name, msg)
 			}
 		})
 
 		uc.users.Del(msg.Prefix.Name)
 
 		if msg.Prefix.Name != uc.nick {
-			uc.forEachDownstream(func(dc *downstreamConn) {
-				dc.SendMessage(msg)
-			})
+			uc.forwardMessage(ctx, msg)
 		}
 	case irc.RPL_TOPIC, irc.RPL_NOTOPIC:
 		var name, topic string
 		if err := parseMessageParams(msg, nil, &name, &topic); err != nil {
 			return err
 		}
-		ch, err := uc.getChannel(name)
-		if err != nil {
-			return err
-		}
-		if msg.Command == irc.RPL_TOPIC {
-			ch.Topic = topic
+		ch := uc.channels.Get(name)
+		if ch == nil {
+			uc.forwardMsgByID(ctx, downstreamID, msg)
 		} else {
-			ch.Topic = ""
+			if msg.Command == irc.RPL_TOPIC {
+				ch.Topic = topic
+			} else {
+				ch.Topic = ""
+			}
 		}
 	case "TOPIC":
 		var name string
@@ -1308,7 +1378,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 		} else {
 			ch.Topic = ""
 		}
-		uc.produce(ch.Name, msg, 0)
+		uc.produce(ctx, ch.Name, msg, 0)
 	case "MODE":
 		var name, modeStr string
 		if err := parseMessageParams(msg, &name, &modeStr); err != nil {
@@ -1324,13 +1394,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 				return err
 			}
 
-			uc.forEachDownstream(func(dc *downstreamConn) {
-				if dc.upstream() == nil {
-					return
-				}
-
-				dc.SendMessage(msg)
-			})
+			uc.forwardMessage(ctx, msg)
 		} else { // channel mode change
 			ch, err := uc.getChannel(name)
 			if err != nil {
@@ -1342,13 +1406,11 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 				return err
 			}
 
-			uc.appendLog(ch.Name, msg)
+			uc.appendLog(ctx, ch.Name, msg)
 
 			c := uc.network.channels.Get(name)
 			if c == nil || !c.Detached {
-				uc.forEachDownstream(func(dc *downstreamConn) {
-					dc.SendMessage(msg)
-				})
+				uc.forwardMessage(ctx, msg)
 			}
 		}
 	case irc.RPL_UMODEIS:
@@ -1365,39 +1427,34 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			return err
 		}
 
-		uc.forEachDownstream(func(dc *downstreamConn) {
-			if dc.upstream() == nil {
-				return
-			}
-
-			dc.SendMessage(msg)
-		})
+		uc.forwardMessage(ctx, msg)
 	case irc.RPL_CHANNELMODEIS:
 		var channel string
 		if err := parseMessageParams(msg, nil, &channel); err != nil {
 			return err
 		}
 		modeStr := ""
+		var modeArgs []string
 		if len(msg.Params) > 2 {
 			modeStr = msg.Params[2]
+			modeArgs = msg.Params[3:]
 		}
 
-		ch, err := uc.getChannel(channel)
-		if err != nil {
-			return err
+		ch := uc.channels.Get(channel)
+		if ch == nil {
+			uc.forwardMsgByID(ctx, downstreamID, msg)
+			return nil
 		}
 
 		firstMode := ch.modes == nil
 		ch.modes = make(map[byte]string)
-		if err := applyChannelModes(ch, modeStr, msg.Params[3:]); err != nil {
+		if err := applyChannelModes(ch, modeStr, modeArgs); err != nil {
 			return err
 		}
 
 		c := uc.network.channels.Get(channel)
 		if firstMode && (c == nil || !c.Detached) {
-			uc.forEachDownstream(func(dc *downstreamConn) {
-				dc.SendMessage(msg)
-			})
+			uc.forwardMessage(ctx, msg)
 		}
 	case xirc.RPL_CREATIONTIME:
 		var channel, creationTime string
@@ -1405,9 +1462,10 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			return err
 		}
 
-		ch, err := uc.getChannel(channel)
-		if err != nil {
-			return err
+		ch := uc.channels.Get(channel)
+		if ch == nil {
+			uc.forwardMsgByID(ctx, downstreamID, msg)
+			return nil
 		}
 
 		firstCreationTime := ch.creationTime == ""
@@ -1415,9 +1473,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 
 		c := uc.network.channels.Get(channel)
 		if firstCreationTime && (c == nil || !c.Detached) {
-			uc.forEachDownstream(func(dc *downstreamConn) {
-				dc.SendMessage(msg)
-			})
+			uc.forwardMessage(ctx, msg)
 		}
 	case xirc.RPL_TOPICWHOTIME:
 		var channel, who, timeStr string
@@ -1425,9 +1481,10 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			return err
 		}
 
-		ch, err := uc.getChannel(channel)
-		if err != nil {
-			return err
+		ch := uc.channels.Get(channel)
+		if ch == nil {
+			uc.forwardMsgByID(ctx, downstreamID, msg)
+			return nil
 		}
 
 		firstTopicWhoTime := ch.TopicWho == nil
@@ -1440,11 +1497,9 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 
 		c := uc.network.channels.Get(channel)
 		if firstTopicWhoTime && (c == nil || !c.Detached) {
-			uc.forEachDownstream(func(dc *downstreamConn) {
-				dc.SendMessage(msg)
-			})
+			uc.forwardMessage(ctx, msg)
 		}
-	case irc.RPL_LIST:
+	case irc.RPL_LISTSTART, irc.RPL_LIST:
 		dc, cmd := uc.currentPendingCommand("LIST")
 		if cmd == nil {
 			return fmt.Errorf("unexpected RPL_LIST: no matching pending LIST")
@@ -1452,7 +1507,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			return nil
 		}
 
-		dc.SendMessage(msg)
+		dc.SendMessage(ctx, msg)
 	case irc.RPL_LISTEND:
 		dc, cmd := uc.dequeueCommand("LIST")
 		if cmd == nil {
@@ -1461,7 +1516,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			return nil
 		}
 
-		dc.SendMessage(msg)
+		dc.SendMessage(ctx, msg)
 	case irc.RPL_NAMREPLY:
 		var name, statusStr, members string
 		if err := parseMessageParams(msg, nil, &statusStr, &name, &members); err != nil {
@@ -1471,7 +1526,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 		ch := uc.channels.Get(name)
 		if ch == nil {
 			// NAMES on a channel we have not joined, forward to downstream
-			uc.forwardMsgByID(downstreamID, msg)
+			uc.forwardMsgByID(ctx, downstreamID, msg)
 			return nil
 		}
 
@@ -1494,7 +1549,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 		ch := uc.channels.Get(name)
 		if ch == nil {
 			// NAMES on a channel we have not joined, forward to downstream
-			uc.forwardMsgByID(downstreamID, msg)
+			uc.forwardMsgByID(ctx, downstreamID, msg)
 			return nil
 		}
 
@@ -1528,7 +1583,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 		}
 		realname := parts[1]
 
-		dc.SendMessage(msg)
+		dc.SendMessage(ctx, msg)
 
 		if uc.shouldCacheUserInfo(nick) {
 			uc.cacheUserInfo(nick, &upstreamUser{
@@ -1536,7 +1591,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 				Hostname: host,
 				Server:   server,
 				Nickname: nick,
-				Flags:    flags,
+				Flags:    stripMemberPrefixes(flags, uc),
 				Realname: realname,
 			})
 		}
@@ -1548,7 +1603,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			return nil
 		}
 
-		dc.SendMessage(msg)
+		dc.SendMessage(ctx, msg)
 
 		if len(cmd.Params) > 1 {
 			fields, _ := xirc.ParseWHOXOptions(cmd.Params[1])
@@ -1559,13 +1614,14 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			if err != nil {
 				return err
 			}
+
 			if uc.shouldCacheUserInfo(info.Nickname) {
 				uc.cacheUserInfo(info.Nickname, &upstreamUser{
 					Nickname: info.Nickname,
 					Username: info.Username,
 					Hostname: info.Hostname,
 					Server:   info.Server,
-					Flags:    info.Flags,
+					Flags:    stripMemberPrefixes(info.Flags, uc),
 					Account:  info.Account,
 					Realname: info.Realname,
 				})
@@ -1581,7 +1637,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			return nil
 		}
 
-		dc.SendMessage(msg)
+		dc.SendMessage(ctx, msg)
 	case xirc.RPL_WHOISCERTFP, xirc.RPL_WHOISREGNICK, irc.RPL_WHOISUSER, irc.RPL_WHOISSERVER, irc.RPL_WHOISCHANNELS, irc.RPL_WHOISOPERATOR, irc.RPL_WHOISIDLE, xirc.RPL_WHOISSPECIAL, xirc.RPL_WHOISACCOUNT, xirc.RPL_WHOISACTUALLY, xirc.RPL_WHOISHOST, xirc.RPL_WHOISMODES, xirc.RPL_WHOISSECURE:
 		dc, cmd := uc.currentPendingCommand("WHOIS")
 		if cmd == nil {
@@ -1590,7 +1646,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			return nil
 		}
 
-		dc.SendMessage(msg)
+		dc.SendMessage(ctx, msg)
 	case irc.RPL_ENDOFWHOIS:
 		dc, cmd := uc.dequeueCommand("WHOIS")
 		if cmd == nil {
@@ -1599,7 +1655,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			return nil
 		}
 
-		dc.SendMessage(msg)
+		dc.SendMessage(ctx, msg)
 	case "INVITE":
 		var nick, channel string
 		if err := parseMessageParams(msg, &nick, &channel); err != nil {
@@ -1608,11 +1664,36 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 
 		weAreInvited := uc.isOurNick(nick)
 
+		blocked := false
+		if weAreInvited {
+			joined := uc.channels.Get(channel) != nil
+			c := uc.network.channels.Get(channel)
+			if !joined && c != nil {
+				// Automatically join a saved channel when we are invited
+				for _, msg := range xirc.GenerateJoin([]string{c.Name}, []string{c.Key}) {
+					uc.SendMessage(ctx, msg)
+				}
+				break
+			}
+
+			sourceCM := uc.network.casemap(msg.Prefix.Name)
+			mt, err := uc.srv.db.GetMessageTarget(ctx, uc.network.ID, sourceCM)
+			if err != nil {
+				uc.logger.Printf("failed to get the message target for %q: %v", sourceCM, err)
+			}
+			if mt != nil && mt.Blocked {
+				blocked = true
+			}
+		}
+
 		uc.forEachDownstream(func(dc *downstreamConn) {
 			if !weAreInvited && !dc.caps.IsEnabled("invite-notify") {
 				return
 			}
-			dc.SendMessage(msg)
+			if blocked && !dc.metadataSubs["soju.im/blocked"] {
+				return
+			}
+			dc.SendMessage(ctx, msg)
 		})
 
 		if weAreInvited {
@@ -1624,7 +1705,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			return err
 		}
 
-		uc.forwardMsgByID(downstreamID, msg)
+		uc.forwardMsgByID(ctx, downstreamID, msg)
 	case irc.RPL_MONONLINE, irc.RPL_MONOFFLINE:
 		var targetsStr string
 		if err := parseMessageParams(msg, nil, &targetsStr); err != nil {
@@ -1662,8 +1743,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			for _, target := range targets {
 				prefix := irc.ParsePrefix(target)
 				if dc.monitored.Has(prefix.Name) {
-					dc.SendMessage(&irc.Message{
-						Prefix:  dc.srv.prefix(),
+					dc.SendMessage(ctx, &irc.Message{
 						Command: msg.Command,
 						Params:  []string{dc.nick, target},
 					})
@@ -1680,8 +1760,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 		uc.forEachDownstream(func(dc *downstreamConn) {
 			for _, target := range targets {
 				if dc.monitored.Has(target) {
-					dc.SendMessage(&irc.Message{
-						Prefix:  dc.srv.prefix(),
+					dc.SendMessage(ctx, &irc.Message{
 						Command: msg.Command,
 						Params:  []string{dc.nick, limit, target},
 					})
@@ -1689,7 +1768,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			}
 		})
 	case irc.RPL_AWAY:
-		uc.forwardMsgByID(downstreamID, msg)
+		uc.forwardMsgByID(ctx, downstreamID, msg)
 	case "AWAY":
 		// Update user flags, if we already have the flags cached
 		uu := uc.users.Get(msg.Prefix.Name)
@@ -1705,9 +1784,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			})
 		}
 
-		uc.forEachDownstream(func(dc *downstreamConn) {
-			dc.SendMessage(msg)
-		})
+		uc.forwardMessage(ctx, msg)
 	case "ACCOUNT":
 		var account string
 		if err := parseMessageParams(msg, &account); err != nil {
@@ -1716,12 +1793,12 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 		uc.cacheUserInfo(msg.Prefix.Name, &upstreamUser{
 			Account: account,
 		})
-		uc.forEachDownstream(func(dc *downstreamConn) {
-			dc.SendMessage(msg)
-		})
+		uc.forwardMessage(ctx, msg)
 	case irc.RPL_BANLIST, irc.RPL_INVITELIST, irc.RPL_EXCEPTLIST, irc.RPL_ENDOFBANLIST, irc.RPL_ENDOFINVITELIST, irc.RPL_ENDOFEXCEPTLIST:
-		uc.forwardMsgByID(downstreamID, msg)
-	case irc.ERR_NOSUCHNICK:
+		uc.forwardMsgByID(ctx, downstreamID, msg)
+	case irc.ERR_NOSUCHNICK, irc.ERR_NOSUCHSERVER:
+		// one argument WHOIS variant errors with NOSUCHNICK
+		// two argument WHOIS variant errors with NOSUCHSERVER
 		var nick, reason string
 		if err := parseMessageParams(msg, nil, &nick, &reason); err != nil {
 			return err
@@ -1732,10 +1809,10 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 		if cmd != nil && cm(cmd.Params[len(cmd.Params)-1]) == cm(nick) {
 			uc.dequeueCommand("WHOIS")
 			if dc != nil {
-				dc.SendMessage(msg)
+				dc.SendMessage(ctx, msg)
 			}
 		} else {
-			uc.forwardMsgByID(downstreamID, msg)
+			uc.forwardMsgByID(ctx, downstreamID, msg)
 		}
 	case xirc.ERR_UNKNOWNERROR, irc.ERR_UNKNOWNCOMMAND, irc.ERR_NEEDMOREPARAMS, irc.RPL_TRYAGAIN:
 		var command, reason string
@@ -1752,7 +1829,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			uc.saslStarted = false
 		}
 
-		uc.forwardMsgByID(downstreamID, msg)
+		uc.forwardMsgByID(ctx, downstreamID, msg)
 	case "FAIL":
 		var command, code string
 		if err := parseMessageParams(msg, &command, &code); err != nil {
@@ -1767,7 +1844,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			downstreamID = dc.id
 		}
 
-		uc.forwardMsgByID(downstreamID, msg)
+		uc.forwardMsgByID(ctx, downstreamID, msg)
 	case "ACK":
 		// Ignore
 	case irc.RPL_NOWAWAY, irc.RPL_UNAWAY:
@@ -1787,9 +1864,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 			return nil
 		}
 
-		uc.forwardMsgByID(downstreamID, msg)
-	case irc.RPL_LISTSTART:
-		// Ignore
+		uc.forwardMsgByID(ctx, downstreamID, msg)
 	case "ERROR":
 		var text string
 		if err := parseMessageParams(msg, &text); err != nil {
@@ -1827,10 +1902,10 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 		if !uc.registered {
 			return registrationError{msg}
 		}
-		uc.forwardMsgByID(downstreamID, msg)
+		uc.forwardMsgByID(ctx, downstreamID, msg)
 	default:
-		uc.logger.Printf("unhandled message: %v", msg)
-		uc.forwardMsgByID(downstreamID, msg)
+		uc.logger.Debugf("unhandled message: %v", msg)
+		uc.forwardMsgByID(ctx, downstreamID, msg)
 	}
 	return nil
 }
@@ -1838,7 +1913,7 @@ func (uc *upstreamConn) handleMessage(ctx context.Context, msg *irc.Message) err
 func (uc *upstreamConn) handleDetachedMessage(ctx context.Context, ch *database.Channel, msg *irc.Message) {
 	if uc.network.detachedMessageNeedsRelay(ch, msg) {
 		uc.forEachDownstream(func(dc *downstreamConn) {
-			dc.relayDetachedMessage(uc.network, msg)
+			dc.relayDetachedMessage(ctx, uc.network, msg)
 		})
 	}
 	if ch.ReattachOn == database.FilterMessage || (ch.ReattachOn == database.FilterHighlight && uc.network.isHighlight(msg)) {
@@ -1938,8 +2013,8 @@ func (uc *upstreamConn) supportsSASL(mech string) bool {
 	}
 
 	mechanisms := strings.Split(v, ",")
-	for _, mech := range mechanisms {
-		if strings.EqualFold(mech, mech) {
+	for _, m := range mechanisms {
+		if strings.EqualFold(m, mech) {
 			return true
 		}
 	}
@@ -2060,6 +2135,18 @@ func (uc *upstreamConn) runUntilRegistered(ctx context.Context) error {
 		}
 	}
 
+	// Connect commands usually send a message to NickServ or similar, which
+	// asynchronously informs the IRC server that the user is logged in. Some
+	// channels require the user to be logged in to be joined. Work around
+	// these old networks by leaving 3 seconds for the bot to react.
+	if len(uc.network.ConnectCommands) > 0 {
+		uc.logger.Debugf("sent %d connect command(s)", len(uc.network.ConnectCommands))
+		select {
+		case <-ctx.Done():
+		case <-time.After(3 * time.Second):
+		}
+	}
+
 	return nil
 }
 
@@ -2103,14 +2190,14 @@ func (uc *upstreamConn) SendMessageLabeled(ctx context.Context, downstreamID uin
 //
 // The internal message ID is returned. If the message isn't recorded in the
 // log file, an empty string is returned.
-func (uc *upstreamConn) appendLog(entity string, msg *irc.Message) (msgID string) {
+func (uc *upstreamConn) appendLog(ctx context.Context, entity string, msg *irc.Message) (msgID string) {
 	if uc.user.msgStore == nil {
 		return ""
 	}
 	if msg.Command == "TAGMSG" {
 		store := false
-		for tag := range storableMessageTags {
-			if _, ok := msg.Tags[tag]; ok {
+		for tag := range msg.Tags {
+			if storableMessageTags[tag] {
 				store = true
 				break
 			}
@@ -2132,11 +2219,11 @@ func (uc *upstreamConn) appendLog(entity string, msg *irc.Message) (msgID string
 		return ""
 	}
 
-	if !uc.network.delivered.HasTarget(entity) {
+	if !uc.network.delivered.Empty() && !uc.network.delivered.HasTarget(entity) {
 		// This is the first message we receive from this target. Save the last
 		// message ID in delivery receipts, so that we can send the new message
 		// in the backlog if an offline client reconnects.
-		lastID, err := uc.user.msgStore.LastMsgID(&uc.network.Network, entityCM, time.Now())
+		lastID, err := uc.user.msgStore.LastMsgID(ctx, &uc.network.Network, entityCM, time.Now())
 		if err != nil {
 			uc.logger.Printf("failed to log message: failed to get last message ID: %v", err)
 			return ""
@@ -2147,7 +2234,7 @@ func (uc *upstreamConn) appendLog(entity string, msg *irc.Message) (msgID string
 		})
 	}
 
-	msgID, err := uc.user.msgStore.Append(&uc.network.Network, entityCM, msg)
+	msgID, err := uc.user.msgStore.Append(ctx, &uc.network.Network, entityCM, msg)
 	if err != nil {
 		uc.logger.Printf("failed to append message to store: %v", err)
 		return ""
@@ -2162,29 +2249,31 @@ func (uc *upstreamConn) appendLog(entity string, msg *irc.Message) (msgID string
 // originID is the id of the downstream (origin) that sent the message. If it is not 0
 // and origin doesn't support echo-message, the message is forwarded to all
 // connections except origin.
-func (uc *upstreamConn) produce(target string, msg *irc.Message, originID uint64) {
+func (uc *upstreamConn) produce(ctx context.Context, target string, msg *irc.Message, originID uint64) {
 	var msgID string
 	if target != "" {
-		msgID = uc.appendLog(target, msg)
+		msgID = uc.appendLog(ctx, target, msg)
 	}
 
 	// Don't forward messages if it's a detached channel
 	ch := uc.network.channels.Get(target)
 	detached := ch != nil && ch.Detached
+	blocked, _ := ctx.Value(contextSourceBlockedKey).(bool)
 
 	uc.forEachDownstream(func(dc *downstreamConn) {
+		if blocked && !dc.metadataSubs["soju.im/blocked"] {
+			return
+		}
 		echo := dc.id == originID && msg.Prefix != nil && uc.isOurNick(msg.Prefix.Name)
 		if !detached && (!echo || dc.caps.IsEnabled("echo-message")) {
-			dc.sendMessageWithID(msg, msgID)
+			dc.sendMessageWithID(ctx, msg, msgID)
 		} else {
-			dc.advanceMessageWithID(msg, msgID)
+			dc.advanceMessageWithID(ctx, msg, msgID)
 		}
 	})
 }
 
-func (uc *upstreamConn) updateAway() {
-	ctx := context.TODO()
-
+func (uc *upstreamConn) updateAway(ctx context.Context) {
 	if !uc.network.AutoAway {
 		return
 	}
@@ -2199,9 +2288,13 @@ func (uc *upstreamConn) updateAway() {
 		return
 	}
 	if away {
+		reason := "Auto away"
+		if uc.caps.IsAvailable("draft/pre-away") {
+			reason = "*"
+		}
 		uc.SendMessage(ctx, &irc.Message{
 			Command: "AWAY",
-			Params:  []string{"Auto away"},
+			Params:  []string{reason},
 		})
 	} else {
 		uc.SendMessage(ctx, &irc.Message{
@@ -2223,13 +2316,11 @@ func (uc *upstreamConn) updateChannelAutoDetach(name string) {
 	uch.updateAutoDetach(ch.DetachAfter)
 }
 
-func (uc *upstreamConn) updateMonitor() {
+func (uc *upstreamConn) updateMonitor(ctx context.Context) {
 	if _, ok := uc.isupport["MONITOR"]; !ok {
 		return
 	}
 
-	ctx := context.TODO()
-
 	add := make(map[string]struct{})
 	var addList []string
 	seen := make(map[string]struct{})
@@ -2329,9 +2420,7 @@ func (uc *upstreamConn) startRegainNickTimer() {
 	})
 }
 
-func (uc *upstreamConn) tryRegainNick(nick string) {
-	ctx := context.TODO()
-
+func (uc *upstreamConn) tryRegainNick(ctx context.Context, nick string) {
 	if uc.regainNickTimer == nil {
 		return
 	}
@@ -2439,3 +2528,31 @@ func (uc *upstreamConn) shouldCacheUserInfo(nick string) bool {
 	})
 	return found
 }
+
+// resolveIPAddr replaces the standard library's DNS resolver to randomize the
+// result order instead of always returning the same IP address. The bouncer
+// will often have bursts of connections to the same host (e.g. on startup) so
+// it's more important for our use-case to distribute the traffic among
+// available IP addresses than to find the fastest link.
+//
+// See: https://todo.sr.ht/~emersion/soju/221
+func resolveIPAddr(ctx context.Context, host string) (*net.IPAddr, error) {
+	ipAddrs, err := net.DefaultResolver.LookupIPAddr(ctx, host)
+	if err != nil {
+		return nil, err
+	}
+
+	// Prefer IPv6 if available, for per-user local IP addresses
+	ip6Addrs := make([]net.IPAddr, 0, len(ipAddrs))
+	for _, ipAddr := range ipAddrs {
+		if ipAddr.IP.To4() == nil {
+			ip6Addrs = append(ip6Addrs, ipAddr)
+		}
+	}
+	if len(ip6Addrs) > 0 {
+		ipAddrs = ip6Addrs
+	}
+
+	i := rand.Intn(len(ipAddrs))
+	return &ipAddrs[i], nil
+}
