@@ -11,16 +11,16 @@ import (
 	"net"
 	"sort"
 	"strings"
+	"sync"
 	"sync/atomic"
 	"time"
 
-	"git.sr.ht/~emersion/soju/xirc"
-
 	"github.com/SherClockHolmes/webpush-go"
 	"gopkg.in/irc.v4"
 
-	"git.sr.ht/~emersion/soju/database"
-	"git.sr.ht/~emersion/soju/msgstore"
+	"codeberg.org/emersion/soju/database"
+	"codeberg.org/emersion/soju/msgstore"
+	"codeberg.org/emersion/soju/xirc"
 )
 
 type UserUpdateFunc func(record *database.User) error
@@ -75,10 +75,11 @@ type eventBroadcast struct {
 type eventStop struct{}
 
 type eventUserUpdate struct {
-	password *string
-	admin    *bool
-	enabled  *bool
-	done     chan error
+	password    *string
+	admin       *bool
+	enabled     *bool
+	maxNetworks *int
+	done        chan error
 }
 
 type eventTryRegainNick struct {
@@ -88,8 +89,7 @@ type eventTryRegainNick struct {
 
 type eventUserRun struct {
 	params []string
-	print  chan string
-	ret    chan error
+	ch     chan userRunMsg
 }
 
 type deliveredClientMap map[string]string // client name -> msg ID
@@ -102,6 +102,10 @@ func newDeliveredStore(cm xirc.CaseMapping) deliveredStore {
 	return deliveredStore{xirc.NewCaseMappingMap[deliveredClientMap](cm)}
 }
 
+func (ds deliveredStore) Empty() bool {
+	return ds.m.Len() == 0
+}
+
 func (ds deliveredStore) HasTarget(target string) bool {
 	return ds.m.Get(target) != nil
 }
@@ -148,12 +152,14 @@ type network struct {
 	logger  Logger
 	stopped chan struct{}
 
-	conn        *upstreamConn
-	channels    xirc.CaseMappingMap[*database.Channel]
-	delivered   deliveredStore
-	pushTargets xirc.CaseMappingMap[time.Time]
-	lastError   error
-	casemap     xirc.CaseMapping
+	conn      *upstreamConn
+	channels  xirc.CaseMappingMap[*database.Channel]
+	delivered deliveredStore
+	lastError error
+	casemap   xirc.CaseMapping
+
+	pushTargetsMutex sync.Mutex
+	pushTargets      xirc.CaseMappingMap[time.Time]
 }
 
 func newNetwork(user *user, record *database.Network, channels []database.Channel) *network {
@@ -218,9 +224,11 @@ func (net *network) runConn(ctx context.Context) error {
 	net.user.srv.metrics.upstreams.Add(1)
 	defer net.user.srv.metrics.upstreams.Add(-1)
 
-	done := ctx.Done()
-	ctx, cancel := context.WithTimeout(ctx, time.Minute)
+	ctx, cancel := context.WithCancel(ctx)
 	defer cancel()
+	done := ctx.Done() // This must not be subject to timeout
+	ctx, cancelTimeout := context.WithTimeout(ctx, time.Minute)
+	defer cancelTimeout()
 
 	uc, err := connectToUpstream(ctx, net)
 	if err != nil {
@@ -311,7 +319,7 @@ func (net *network) stop() {
 	}
 }
 
-func (net *network) detach(ch *database.Channel) {
+func (net *network) detach(ctx context.Context, ch *database.Channel) {
 	if ch.Detached {
 		return
 	}
@@ -322,7 +330,7 @@ func (net *network) detach(ch *database.Channel) {
 
 	if net.user.msgStore != nil {
 		nameCM := net.casemap(ch.Name)
-		lastID, err := net.user.msgStore.LastMsgID(&net.Network, nameCM, time.Now())
+		lastID, err := net.user.msgStore.LastMsgID(ctx, &net.Network, nameCM, time.Now())
 		if err != nil {
 			net.logger.Printf("failed to get last message ID for channel %q: %v", ch.Name, err)
 		}
@@ -337,7 +345,7 @@ func (net *network) detach(ch *database.Channel) {
 	}
 
 	net.forEachDownstream(func(dc *downstreamConn) {
-		dc.SendMessage(&irc.Message{
+		dc.SendMessage(ctx, &irc.Message{
 			Prefix:  dc.prefix(),
 			Command: "PART",
 			Params:  []string{ch.Name, "Detach"},
@@ -364,7 +372,7 @@ func (net *network) attach(ctx context.Context, ch *database.Channel) {
 	}
 
 	net.forEachDownstream(func(dc *downstreamConn) {
-		dc.SendMessage(&irc.Message{
+		dc.SendMessage(ctx, &irc.Message{
 			Prefix:  dc.prefix(),
 			Command: "JOIN",
 			Params:  []string{ch.Name},
@@ -403,7 +411,9 @@ func (net *network) updateCasemapping(newCasemap xirc.CaseMapping) {
 	net.casemap = newCasemap
 	net.channels.SetCaseMapping(newCasemap)
 	net.delivered.m.SetCaseMapping(newCasemap)
+	net.pushTargetsMutex.Lock()
 	net.pushTargets.SetCaseMapping(newCasemap)
+	net.pushTargetsMutex.Unlock()
 	if uc := net.conn; uc != nil {
 		uc.channels.SetCaseMapping(newCasemap)
 		uc.channels.ForEach(func(_ string, uch *upstreamChannel) {
@@ -537,9 +547,9 @@ func newUser(srv *Server, record *database.User) *user {
 	logger := &prefixLogger{srv.Logger, fmt.Sprintf("user %q: ", record.Username)}
 
 	var msgStore msgstore.Store
-	switch srv.Config().LogDriver {
+	switch srv.Config().MsgStoreDriver {
 	case "fs":
-		msgStore = msgstore.NewFSStore(srv.Config().LogPath, record)
+		msgStore = msgstore.NewFSStore(srv.Config().MsgStorePath, record)
 	case "db":
 		msgStore = msgstore.NewDBStore(srv.db)
 	case "memory":
@@ -596,7 +606,13 @@ func (u *user) run() {
 		close(u.done)
 	}()
 
-	networks, err := u.srv.db.ListNetworks(context.TODO(), u.ID)
+	ctx, cancel := context.WithCancel(context.TODO())
+	go func() {
+		<-u.done
+		cancel()
+	}()
+
+	networks, err := u.srv.db.ListNetworks(ctx, u.ID)
 	if err != nil {
 		u.logger.Printf("failed to list networks for user %q: %v", u.Username, err)
 		return
@@ -608,7 +624,7 @@ func (u *user) run() {
 
 	for _, record := range networks {
 		record := record
-		channels, err := u.srv.db.ListChannels(context.TODO(), record.ID)
+		channels, err := u.srv.db.ListChannels(ctx, record.ID)
 		if err != nil {
 			u.logger.Printf("failed to list channels for user %q, network %q: %v", u.Username, record.GetName(), err)
 			continue
@@ -618,7 +634,7 @@ func (u *user) run() {
 		u.networks = append(u.networks, network)
 
 		if u.hasPersistentMsgStore() {
-			receipts, err := u.srv.db.ListDeliveryReceipts(context.TODO(), record.ID)
+			receipts, err := u.srv.db.ListDeliveryReceipts(ctx, record.ID)
 			if err != nil {
 				u.logger.Printf("failed to load delivery receipts for user %q, network %q: %v", u.Username, network.GetName(), err)
 				return
@@ -633,35 +649,37 @@ func (u *user) run() {
 	}
 
 	for e := range u.events {
+		// TODO: set a timeout for event handling
+
 		switch e := e.(type) {
 		case eventUpstreamConnected:
 			uc := e.uc
 
 			uc.network.conn = uc
 
-			uc.updateAway()
-			uc.updateMonitor()
+			uc.updateAway(ctx)
+			uc.updateMonitor(ctx)
 
 			uc.forEachDownstream(func(dc *downstreamConn) {
-				dc.updateSupportedCaps()
+				dc.updateSupportedCaps(ctx)
 
 				if !dc.caps.IsEnabled("soju.im/bouncer-networks") {
-					sendServiceNOTICE(dc, fmt.Sprintf("connected to %s", uc.network.GetName()))
+					sendServiceNOTICE(ctx, dc, fmt.Sprintf("connected to %s", uc.network.GetName()))
 				}
 
-				dc.updateNick()
-				dc.updateHost()
-				dc.updateRealname()
-				dc.updateAccount()
+				dc.updateNick(ctx)
+				dc.updateHost(ctx)
+				dc.updateRealname(ctx)
+				dc.updateAccount(ctx)
 				dc.updateCasemapping()
 			})
-			u.notifyBouncerNetworkState(uc.network.ID, irc.Tags{
+			u.notifyBouncerNetworkState(ctx, uc.network.ID, irc.Tags{
 				"state": "connected",
 				"error": "",
 			})
 			uc.network.lastError = nil
 		case eventUpstreamDisconnected:
-			u.handleUpstreamDisconnected(e.uc)
+			u.handleUpstreamDisconnected(ctx, e.uc)
 		case eventUpstreamConnectionError:
 			net := e.net
 
@@ -674,21 +692,21 @@ func (u *user) run() {
 
 			if !stopped && (net.lastError == nil || net.lastError.Error() != e.err.Error()) {
 				net.forEachDownstream(func(dc *downstreamConn) {
-					sendServiceNOTICE(dc, fmt.Sprintf("failed connecting/registering to %s: %v", net.GetName(), e.err))
+					sendServiceNOTICE(ctx, dc, fmt.Sprintf("failed connecting/registering to %s: %v", net.GetName(), e.err))
 				})
 			}
 			net.lastError = e.err
-			u.notifyBouncerNetworkState(net.ID, irc.Tags{
+			u.notifyBouncerNetworkState(ctx, net.ID, irc.Tags{
 				"error": net.lastError.Error(),
 			})
 		case eventUpstreamError:
 			uc := e.uc
 
 			uc.forEachDownstream(func(dc *downstreamConn) {
-				sendServiceNOTICE(dc, fmt.Sprintf("disconnected from %s: %v", uc.network.GetName(), e.err))
+				sendServiceNOTICE(ctx, dc, fmt.Sprintf("disconnected from %s: %v", uc.network.GetName(), e.err))
 			})
 			uc.network.lastError = e.err
-			u.notifyBouncerNetworkState(uc.network.ID, irc.Tags{
+			u.notifyBouncerNetworkState(ctx, uc.network.ID, irc.Tags{
 				"error": uc.network.lastError.Error(),
 			})
 		case eventUpstreamMessage:
@@ -697,7 +715,7 @@ func (u *user) run() {
 				uc.logger.Printf("ignoring message on closed connection: %v", msg)
 				break
 			}
-			if err := uc.handleMessage(context.TODO(), msg); err != nil {
+			if err := uc.handleMessage(ctx, msg); err != nil {
 				uc.logger.Printf("failed to handle message %q: %v", msg, err)
 			}
 		case eventChannelDetach:
@@ -706,13 +724,12 @@ func (u *user) run() {
 			if c == nil || c.Detached {
 				continue
 			}
-			uc.network.detach(c)
-			if err := uc.srv.db.StoreChannel(context.TODO(), uc.network.ID, c); err != nil {
+			uc.network.detach(ctx, c)
+			if err := uc.srv.db.StoreChannel(ctx, uc.network.ID, c); err != nil {
 				u.logger.Printf("failed to store updated detached channel %q: %v", c.Name, err)
 			}
 		case eventDownstreamConnected:
 			dc := e.dc
-			ctx := context.TODO()
 
 			if dc.network != nil {
 				dc.monitored.SetCaseMapping(dc.network.casemap)
@@ -729,7 +746,7 @@ func (u *user) run() {
 			}
 
 			if !u.Enabled {
-				dc.SendMessage(&irc.Message{
+				dc.SendMessage(ctx, &irc.Message{
 					Command: "ERROR",
 					Params:  []string{"This bouncer account is disabled"},
 				})
@@ -737,13 +754,13 @@ func (u *user) run() {
 				break
 			}
 
-			if err := dc.welcome(ctx); err != nil {
+			if err := dc.welcome(ctx, u); err != nil {
 				if ircErr, ok := err.(ircError); ok {
 					msg := ircErr.Message.Copy()
 					msg.Prefix = dc.srv.prefix()
-					dc.SendMessage(msg)
+					dc.SendMessage(ctx, msg)
 				} else {
-					dc.SendMessage(&irc.Message{
+					dc.SendMessage(ctx, &irc.Message{
 						Command: "ERROR",
 						Params:  []string{"Internal server error"},
 					})
@@ -756,20 +773,19 @@ func (u *user) run() {
 			u.downstreamConns = append(u.downstreamConns, dc)
 			u.numDownstreamConns.Add(1)
 
-			dc.forEachNetwork(func(network *network) {
-				if network.lastError != nil {
-					sendServiceNOTICE(dc, fmt.Sprintf("disconnected from %s: %v", network.GetName(), network.lastError))
-				}
-			})
+			if network := dc.network; network != nil && network.lastError != nil {
+				sendServiceNOTICE(ctx, dc, fmt.Sprintf("disconnected from %s: %v", network.GetName(), network.lastError))
+			}
 
 			u.forEachUpstream(func(uc *upstreamConn) {
-				uc.updateAway()
+				uc.updateAway(ctx)
 			})
 
-			u.bumpDownstreamInteractionTime(ctx)
+			if !dc.impersonating {
+				u.bumpDownstreamInteractionTime(ctx)
+			}
 		case eventDownstreamDisconnected:
 			dc := e.dc
-			ctx := context.TODO()
 
 			for i := range u.downstreamConns {
 				if u.downstreamConns[i] == dc {
@@ -779,38 +795,36 @@ func (u *user) run() {
 				}
 			}
 
-			dc.forEachNetwork(func(net *network) {
-				net.storeClientDeliveryReceipts(ctx, dc.clientName)
-			})
+			if dc.network != nil {
+				dc.network.storeClientDeliveryReceipts(ctx, dc.clientName)
+			}
 
 			u.forEachUpstream(func(uc *upstreamConn) {
 				uc.cancelPendingCommandsByDownstreamID(dc.id)
-				uc.updateAway()
-				uc.updateMonitor()
+				uc.updateAway(ctx)
+				uc.updateMonitor(ctx)
 			})
 
-			u.bumpDownstreamInteractionTime(ctx)
+			if !dc.impersonating {
+				u.bumpDownstreamInteractionTime(ctx)
+			}
 		case eventDownstreamMessage:
 			msg, dc := e.msg, e.dc
 			if dc.isClosed() {
 				dc.logger.Printf("ignoring message on closed connection: %v", msg)
 				break
 			}
-			err := dc.handleMessage(context.TODO(), msg)
-			if ircErr, ok := err.(ircError); ok {
-				ircErr.Message.Prefix = dc.srv.prefix()
-				dc.SendMessage(ircErr.Message)
-			} else if err != nil {
+			if err := dc.handleMessage(ctx, msg); err != nil {
 				dc.logger.Printf("failed to handle message %q: %v", msg, err)
 				dc.Close()
 			}
 		case eventBroadcast:
 			msg := e.msg
 			for _, dc := range u.downstreamConns {
-				dc.SendMessage(msg)
+				dc.SendMessage(ctx, msg)
 			}
 		case eventUserUpdate:
-			e.done <- u.updateUser(context.TODO(), func(record *database.User) error {
+			e.done <- u.updateUser(ctx, func(record *database.User) error {
 				if e.password != nil {
 					record.Password = *e.password
 				}
@@ -820,6 +834,9 @@ func (u *user) run() {
 				if e.enabled != nil {
 					record.Enabled = *e.enabled
 				}
+				if e.maxNetworks != nil {
+					record.MaxNetworks = *e.maxNetworks
+				}
 				return nil
 			})
 
@@ -831,27 +848,31 @@ func (u *user) run() {
 				}
 			}
 		case eventTryRegainNick:
-			e.uc.tryRegainNick(e.nick)
+			e.uc.tryRegainNick(ctx, e.nick)
 		case eventUserRun:
-			ctx := context.TODO()
 			err := handleServiceCommand(&serviceContext{
 				Context: ctx,
 				user:    u,
 				srv:     u.srv,
-				admin:   u.Admin,
+				// Here we are setting an admin context on any user run command
+				// that is run.
+				// As a reminder, user run can only be run by an admin.
+				// This enables admins to run actions on a user with admin rights,
+				// for example to add a network past the user limit.
+				// Non-admin users cannot run user run, not even on themselves,
+				// so this cannot be used to escalate privileges.
+				admin: true,
 				print: func(text string) {
 					// Avoid blocking on e.print in case our context is canceled.
-					// This is a no-op right now because we use context.TODO(),
-					// but might be useful later when we add timeouts.
 					select {
 					case <-ctx.Done():
-					case e.print <- text:
+					case e.ch <- userRunMsg{message: text}:
 					}
 				},
 			}, e.params)
 			select {
 			case <-ctx.Done():
-			case e.ret <- err:
+			case e.ch <- userRunMsg{err: err}:
 			}
 		case eventStop:
 			for _, dc := range u.downstreamConns {
@@ -861,7 +882,7 @@ func (u *user) run() {
 				n.stop()
 
 				n.delivered.ForEachClient(func(clientName string) {
-					n.storeClientDeliveryReceipts(context.TODO(), clientName)
+					n.storeClientDeliveryReceipts(ctx, clientName)
 				})
 			}
 			return
@@ -871,7 +892,7 @@ func (u *user) run() {
 	}
 }
 
-func (u *user) handleUpstreamDisconnected(uc *upstreamConn) {
+func (u *user) handleUpstreamDisconnected(ctx context.Context, uc *upstreamConn) {
 	uc.network.conn = nil
 
 	uc.stopRegainNickTimer()
@@ -882,7 +903,7 @@ func (u *user) handleUpstreamDisconnected(uc *upstreamConn) {
 	})
 
 	uc.forEachDownstream(func(dc *downstreamConn) {
-		dc.updateSupportedCaps()
+		dc.updateSupportedCaps(ctx)
 	})
 
 	// If the network has been removed, don't send a state change notification
@@ -897,23 +918,34 @@ func (u *user) handleUpstreamDisconnected(uc *upstreamConn) {
 		return
 	}
 
-	u.notifyBouncerNetworkState(uc.network.ID, irc.Tags{"state": "disconnected"})
+	u.notifyBouncerNetworkState(ctx, uc.network.ID, irc.Tags{"state": "disconnected"})
 
 	if uc.network.lastError == nil {
 		uc.forEachDownstream(func(dc *downstreamConn) {
 			if !dc.caps.IsEnabled("soju.im/bouncer-networks") {
-				sendServiceNOTICE(dc, fmt.Sprintf("disconnected from %s", uc.network.GetName()))
+				sendServiceNOTICE(ctx, dc, fmt.Sprintf("disconnected from %s", uc.network.GetName()))
 			}
 		})
 	}
 }
 
-func (u *user) notifyBouncerNetworkState(netID int64, attrs irc.Tags) {
+func (u *user) notifyBouncerNetworkState(ctx context.Context, netID int64, attrs irc.Tags) {
+	// Don't send state updates for removed networks
+	found := false
+	for _, net := range u.networks {
+		if net.ID == netID {
+			found = true
+			break
+		}
+	}
+	if !found {
+		return
+	}
+
 	netIDStr := fmt.Sprintf("%v", netID)
 	for _, dc := range u.downstreamConns {
 		if dc.caps.IsEnabled("soju.im/bouncer-networks-notify") {
-			dc.SendMessage(&irc.Message{
-				Prefix:  dc.srv.prefix(),
+			dc.SendMessage(ctx, &irc.Message{
 				Command: "BOUNCER",
 				Params:  []string{"NETWORK", netIDStr, attrs.String()},
 			})
@@ -950,6 +982,26 @@ func (u *user) removeNetwork(network *network) {
 	panic("tried to remove a non-existing network")
 }
 
+func (u *user) canEnableNewNetwork() error {
+	max := u.MaxNetworks
+	if max < 0 {
+		max = u.srv.Config().MaxUserNetworks
+	}
+	if max < 0 {
+		return nil
+	}
+	n := 0
+	for _, network := range u.networks {
+		if network.Enabled {
+			n++
+		}
+	}
+	if n >= max {
+		return fmt.Errorf("maximum number of enabled networks reached")
+	}
+	return nil
+}
+
 func (u *user) checkNetwork(record *database.Network) error {
 	url, err := record.URL()
 	if err != nil {
@@ -973,6 +1025,9 @@ func (u *user) checkNetwork(record *database.Network) error {
 			return fmt.Errorf("%v:// URL must not have a path", url.Scheme)
 		}
 	case "irc+unix", "unix":
+		if !u.Admin {
+			return fmt.Errorf("non-admin users cannot add networks with a unix address")
+		}
 		if url.Host != "" {
 			return fmt.Errorf("%v:// URL must not have a host", url.Scheme)
 		}
@@ -1000,17 +1055,19 @@ func (u *user) checkNetwork(record *database.Network) error {
 	return nil
 }
 
-func (u *user) createNetwork(ctx context.Context, record *database.Network) (*network, error) {
+func (u *user) createNetwork(ctx context.Context, record *database.Network, enforceLimit bool) (*network, error) {
 	if record.ID != 0 {
 		panic("tried creating an already-existing network")
 	}
 
-	if err := u.checkNetwork(record); err != nil {
-		return nil, err
+	if enforceLimit && record.Enabled {
+		if err := u.canEnableNewNetwork(); err != nil {
+			return nil, err
+		}
 	}
 
-	if max := u.srv.Config().MaxUserNetworks; max >= 0 && len(u.networks) >= max {
-		return nil, fmt.Errorf("maximum number of networks reached")
+	if err := u.checkNetwork(record); err != nil {
+		return nil, err
 	}
 
 	network := newNetwork(u, record, nil)
@@ -1022,16 +1079,22 @@ func (u *user) createNetwork(ctx context.Context, record *database.Network) (*ne
 	u.addNetwork(network)
 
 	attrs := getNetworkAttrs(network)
-	u.notifyBouncerNetworkState(network.ID, attrs)
+	u.notifyBouncerNetworkState(ctx, network.ID, attrs)
 
 	return network, nil
 }
 
-func (u *user) updateNetwork(ctx context.Context, record *database.Network) (*network, error) {
+func (u *user) updateNetwork(ctx context.Context, record *database.Network, enforceLimit bool) (*network, error) {
 	if record.ID == 0 {
 		panic("tried updating a new network")
 	}
 
+	if enforceLimit && record.Enabled {
+		if err := u.canEnableNewNetwork(); err != nil {
+			return nil, err
+		}
+	}
+
 	// If the nickname/realname is reset to the default, just wipe the
 	// per-network setting
 	if record.Nick == u.Nick {
@@ -1068,7 +1131,7 @@ func (u *user) updateNetwork(ctx context.Context, record *database.Network) (*ne
 	network.stop()
 	if network.conn != nil {
 		// Note: this will set network.conn to nil
-		u.handleUpstreamDisconnected(network.conn)
+		u.handleUpstreamDisconnected(ctx, network.conn)
 	}
 
 	// Patch downstream connections to use our fresh updated network
@@ -1082,6 +1145,14 @@ func (u *user) updateNetwork(ctx context.Context, record *database.Network) (*ne
 	// otherwise they'll get closed
 	u.removeNetwork(network)
 
+	// Transfer network attributes that aren't in the record
+	updatedNetwork.delivered = network.delivered
+	updatedNetwork.pushTargets = network.pushTargets
+	if updatedNetwork.Enabled {
+		updatedNetwork.lastError = network.lastError
+	}
+	updatedNetwork.updateCasemapping(network.casemap)
+
 	// The filesystem message store needs to be notified whenever the network
 	// is renamed
 	renameNetMsgStore, ok := u.msgStore.(msgstore.RenameNetworkStore)
@@ -1096,7 +1167,7 @@ func (u *user) updateNetwork(ctx context.Context, record *database.Network) (*ne
 
 	// TODO: only broadcast attributes that have changed
 	attrs := getNetworkAttrs(updatedNetwork)
-	u.notifyBouncerNetworkState(updatedNetwork.ID, attrs)
+	u.notifyBouncerNetworkState(ctx, updatedNetwork.ID, attrs)
 
 	return updatedNetwork, nil
 }
@@ -1116,8 +1187,7 @@ func (u *user) deleteNetwork(ctx context.Context, id int64) error {
 	idStr := fmt.Sprintf("%v", network.ID)
 	for _, dc := range u.downstreamConns {
 		if dc.caps.IsEnabled("soju.im/bouncer-networks-notify") {
-			dc.SendMessage(&irc.Message{
-				Prefix:  dc.srv.prefix(),
+			dc.SendMessage(ctx, &irc.Message{
 				Command: "BOUNCER",
 				Params:  []string{"NETWORK", idStr, "*"},
 			})
@@ -1185,7 +1255,7 @@ func (u *user) updateUser(ctx context.Context, update UserUpdateFunc) error {
 
 		var netErr error
 		for _, net := range needUpdate {
-			if _, err := u.updateNetwork(ctx, &net); err != nil {
+			if _, err := u.updateNetwork(ctx, &net, false); err != nil {
 				netErr = err
 			}
 		}
@@ -1238,27 +1308,15 @@ func (u *user) FormatServerTime(t time.Time) string {
 	return xirc.FormatServerTime(t)
 }
 
-// localAddrForHost returns the local address to use when connecting to host.
+// localTCPAddr returns the local address to use when connecting to a host.
 // A nil address is returned when the OS should automatically pick one.
-func (u *user) localTCPAddrForHost(ctx context.Context, host string) (*net.TCPAddr, error) {
+func (u *user) localTCPAddr(remoteIP net.IP) (*net.TCPAddr, error) {
 	upstreamUserIPs := u.srv.Config().UpstreamUserIPs
 	if len(upstreamUserIPs) == 0 {
 		return nil, nil
 	}
 
-	ips, err := net.DefaultResolver.LookupIP(ctx, "ip", host)
-	if err != nil {
-		return nil, err
-	}
-
-	wantIPv6 := false
-	for _, ip := range ips {
-		if ip.To4() == nil {
-			wantIPv6 = true
-			break
-		}
-	}
-
+	wantIPv6 := remoteIP.To4() == nil
 	var ipNet *net.IPNet
 	for _, in := range upstreamUserIPs {
 		if wantIPv6 == (in.IP.To4() == nil) {
